import maya.mel as mel 
import mayaUtils
from maya.api import MMatrix
from mayaUtils import Translate
reload(mayaUtils)
N = None

from luyaPhysicsUtils import *
from pyBullet import *

def btVector3ToMVector( bt ):
    """ Converts a btVectr to an MVector """
    return MVector( bt.getX(), bt.getY(), bt.getZ() )

def MVectorTobtVector3( m ):
    """ Convert an MVector to a btVector3 """
    return btVector3( m[0], m[1], m[2] )

def splitTransform(T):
	"""
		Takes a 4x4 transform matrix and turns it into a position vector and
		3x3 Rotation Matrix.
	"""
	rotation = btMatrix3x3( T[0], T[4], T[8],
	                        T[1], T[5], T[9],
	                        T[2], T[6], T[10]
	            )
	position = btVector3( T[12], T[13], T[14] )
	return (rotation,position)

class DynamicBody:
	def __init__(self,w,shapeName,shape):
		self.w = w
		self.transformIndex = w.transformIndex ; w.transformIndex += 1
		
		## The Maya Object
		transformName = mel.listRelatives(shapeName,p=N,f=N)[0]
		self.transformName = transformName
		
		#Find out if it is a root rigid body and if so, add it to a list
		if len( mel.listRelatives(transformName,p=N,f=N) ) == 0:
			w.roots.append(transformName)

		self.mass = mel.getAttr("%s.mass" % shapeName)

		#Attributes
		R,T = splitTransform( mel.xform( transformName, q=N, ws=N, a=N, matrix=N ) )
		self.body = w.sWorld.RigidBody( btTransform(R,T), shape, self.mass )
		
		## Adding it to the world class
		if self.mass == 0:
			w.passive[transformName] = self
		else:
			w.active[transformName] = self
			
		w.active_passive[transformName] = self
		w.dynamics.__dict__[transformName.replace("|","_")[1:]] = self
		
	def setTransformFromBullet(self):
		T = self.w.transforms[ self.transformIndex ]
		mel.xform( self.transformName, ws=N,matrix=T.getMatrix4x4() )
		
	def setTransformFromMaya(self):
		pass
		
	def isActive(self):
		if self.mass == 0:
			return False
		else:
			return True

class DynamicJoint:
	def __init__(self, w, shapeName, parent, child):
		i = shapeName
		parentBody = w.active_passive[parent].body
		childBody = w.active_passive[child].body
		
		#Get the postion of the joint
		tParent = Translate(i) - Translate(parent)
		tChild = Translate(i) - Translate(child)
		
		tParentV = MVectorTobtVector3( tParent )
		tChildV = MVectorTobtVector3( tChild )
		
		tParentT = btTransform.getIdentity()
		tChildT = btTransform.getIdentity()
		
		tParentT.setOrigin( tParentV )
		tChildT.setOrigin( tChildV )
		
		joint = btGeneric6DofConstraint( parentBody, childBody, tParentT, tChildT )
		
		w.w.addConstraint( joint )
		
		#Packaging it up
		self.joint = joint
		self.transformName = i
		self.posParent = tParent
		self.posChild = tChild
		self.parentName = parent
		self.childName = child
		
		#Adding it to the world class
		w.joints[i] = self
		if parent == None:
			w.roots.append(i)
		w.dynamics.__dict__[i.replace("|","_")[1:]] = self

		
	def setTransformFromBullet(self):
		#print dir(self.joint)
		#pos = self.joint.getCenterOfMassPosition()
		pass

class World:
	def __init__(self, gravity=btVector3(0,-9.8,0), ground=True ):
		self.sWorld = btWorld( gravity )
		self.w = self.sWorld.getWorld()
		self.world = self.w
		
		self.active = {} #Active Rigid Bodies
		self.passive = {} #Active Rigid Bodies
		self.active_passive = {}
		self.roots = [] #Root Rigid Bodies
		self.joints = {} #Joints
		self.collisionSet = {} # Collision Sets
		self.dynamics = atom() #Joints and Rigid Bodies via names
		
		self.transformIndex = 0
		self.transforms = []
		
		"""Parse the maya scene and find all the rigid bodies and joints to be simulated"""
		self.parseScene()
		
		self.cleanup()
		
	def end(self):
		pass
		
	def parseScene( self ):
		"""
			Parses the maya scene to find all rigid bodies and joints.
		"""
		#Looking for capsules
		for shapeName in mel.ls(type="luCapsule",l=N):
			## The Maya Object
			DynamicBody(self,shapeName,["radius","length"])
			
		#Looking for boxes
		for shapeName in mel.ls(type="luBox",l=N):
			## The Maya Object
			height = mel.getAttr(shapeName+".height")
			width = mel.getAttr(shapeName+".width")
			depth = mel.getAttr(shapeName+".depth")
			shape = btBoxShapeNew( btVector3( height, width, depth ) )
			DynamicBody(self, shapeName, shape)
		
		#Looking for joints which have 1 parent and 1 child capsule
		#Select all joints
		for jointShape in mel.ls(type="joint",l=N):
			#Check that it is dynamic, and if it is, get its parent and child
			(isDynamic,parent,child) = isDynamicJoint(jointShape)
			
			if isDynamic and parent != None :
				DynamicJoint( self, jointShape, parent, child )
		
		#Remove any keyframes which might already exist
		self.cleanup()
	
	def cleanup(self):
		for i in self.active.keys():
			attrs=[	"translateX","translateY","translateZ",
					"rotateX","rotateY","rotateZ",
					"scaleX","scaleY","scaleZ",
					"visibility"
			]
			for j in attrs:
				attrname = "%s_%s" % (i,j)
				if mel.objExists(attrname):
					mel.delete(attrname)

	def m_subStep(self,name):
		"""This is an internal function, don't use it.
			It recursively travels down the hierarchy of a rig and keys the relevent transforms.
		"""
		if len(name) > 0:
			objects = self.active
			j = self.joints
			
			#Is it a rigid body
			if self.active.has_key(name) :
				self.active[name].setTransformFromBullet()
				mel.setKeyframe(name)
				
			#Is it a joint

			if j.has_key(name):
				j[name].setTransformFromBullet()
				mel.setKeyframe(name)
			
			children = mel.listRelatives(name,f=N,c=N,type=["transform","joint"])
			for i in children:
				self.m_subStep(i)
		
	def step(self,timestep=0.1):
		# Go through all the passive rigid bodies and update there transforms based on what they're given
		# from Maya.
		self.transforms = self.sWorld.getTransforms()
		
		for i in self.passive.values():
			i.setTransformFromMaya()
				
		for i in self.roots:
			self.m_subStep(i)
			
		self.w.stepSimulation( timestep )
		