import maya.mel as mel 
from maya.api import MGlobal as M
from maya.api import MEulerRotation
from maya.api import MQuaternion
from maya.api import MVector
from maya.api import MSelectionList
from math import *
import ode
N = None

"""
	Thanks to Paul Kinnane on the pyODE forum for the information on the Param ... 3 arguments.
	Genius.
"""
ParamLoStop = ode.ParamLoStop
ParamHiStop = ode.ParamHiStop
ParamVel = ode.ParamVel
ParamFMax = ode.ParamFMax

ParamLoStop2 = ode.ParamLoStop2
ParamHiStop2 = ode.ParamHiStop2
ParamVel2 = ode.ParamVel2
ParamFMax2 = ode.ParamFMax2

ParamLoStop3 = 512+0
ParamHiStop3 = 512+1
ParamVel3 = 512+2
ParamFMax3 = 512+3

class atom:
	pass

def near_callback(w,geom1,geom2):
	contacts = ode.collide(geom1,geom2) 
	
	for i in contacts:
		j = ode.ContactJoint(w.w,w.contacts,i)
		#i.setBounce(0.02)
		i.setMu(5000)
		j.attach( geom1.getBody(), geom2.getBody() )

def isRigidTransform(name):
	"""Given the name of a transform, this returns true if the transform contains """
	shapes = mel.listRelatives(name,f=N,c=N,s=N)
	if len(shapes) > 0:
		nodeType = mel.nodeType(shapes[0])
		if nodeType == "luCapsule":
			return True
		elif nodeType == "luBox":
			return True
	return False

def isDynamicJoint(name):
	"""Given the name of a joint, return true if it is dynamic, also return the name of the parent and child.
		If a joint has no child then it is not dynamic and (False,None,None) is returned.
	"""
	#Get the parent
	parent = mel.listRelatives(name,f=N,p=N)
	#Get the children
	child = mel.listRelatives(name,f=N,c=N)
	
	#Check the joint has a child rigid transform
	for j in child:
		if isRigidTransform( j ) :
			#Check the joint has a parent rigid transform
			if len(parent) > 0 and isRigidTransform( parent[0] ):
				return (True,parent[0],j)
			else:
				return (True,None,j)
	return (False,None,None)

class DynamicBody:
	def __init__(self,shapeName):
		## The Maya Object
		transformName = mel.listRelatives(shapeName,p=N,f=N)[0]
		
		#Find out if it is a root rigid body and if so, add it to a list
		if len( mel.listRelatives(transformName,p=N,f=N) ) == 0:
			self.roots.append(transformName)
			
		#Attributes
		#Transform
		position = mel.xform(transformName,q=N,ws=N,t=N)
		rotation = mel.xform(transformName,q=N,ws=N,ro=N)

		E = MEulerRotation( radians(rotation[0]), radians(rotation[1]), radians(rotation[2]) )
		Q = E.asQuaternion()

		#Material
		density = mel.getAttr("%s.mass" % shapeName)
		
		## The Mass and Physics Body
		if density != 0:
			body = ode.Body(self.w)
			body.setPosition( position )
			body.setQuaternion((Q.w,Q.x,Q.y,Q.z))
		
		## Packaging it up
		self.body = body
		self.name = transformName

class World:
	def __init__(self, gravity=(0,-9.8,0), ground=True ):
		self.w = ode.World()
		self.space = ode.Space()
		self.contacts = ode.JointGroup()
		self.w.setGravity( gravity )
		
		self.objects = {} #Rigid Bodies
		self.roots = [] #Root Rigid Bodies
		self.joints = {} #Joints
		self.dynamics = atom() #Joints and Rigid Bodies via names
		
		if ground : self.ground = ode.GeomPlane(self.space,normal=(0,1,0))
		
		#Parse the maya scene and find all the rigid bodies and joints to be simulated
		self.parseScene()
		
	def end(self):
		pass
		
	def parseScene( self ):
		"""
			Parses the maya scene to find all rigid bodies and joints.
		"""
		#Looking for capsules
		for i in mel.ls(type="luCapsule",l=N):
			## The Maya Object
			shapeName=i
			transformName = mel.listRelatives(shapeName,p=N,f=N)[0]
			
			#Find out if it is a root rigid body and if so, add it to a list
			if len( mel.listRelatives(transformName,p=N,f=N) ) == 0:
				self.roots.append(transformName)
				
			#Attributes
			#Transform
			position = mel.xform(transformName,q=N,ws=N,t=N)
			rotation = mel.xform(transformName,q=N,ws=N,ro=N)

			#Material
			density = mel.getAttr("%s.mass" % shapeName)
			#Shape
			radius = mel.getAttr("%s.radius" % shapeName)
			length = mel.getAttr("%s.length" % shapeName)
			
			## The Mass and Physics Body
			body = ode.Body(self.w)
			if density != 0:
				mass = ode.Mass()
				mass.setCappedCylinder(density,3,radius,length)
				body.setMass(mass)
			body.setPosition( position )
			
			E = MEulerRotation( radians(rotation[0]), radians(rotation[1]), radians(rotation[2]) )
			Q = E.asQuaternion()
			body.setQuaternion((Q.w,Q.x,Q.y,Q.z))
			
			## The Collision Object
			collision = ode.GeomCCylinder(self.space,radius=radius,length=length)
			collision.setBody(body)
			
			## Packaging it up
			a = atom()
			a.collision = collision
			a.body = body
			if density != 0:
				a.mass = mass
			a.name = transformName
			## Adding it to the world class
			self.objects[transformName] = a
			self.dynamics.__dict__[transformName.replace("|","_")[1:]] = a

		#Looking for boxes
		for i in mel.ls(type="luBox",l=N):
			## The Maya Object
			shapeName=i
			transformName = mel.listRelatives(shapeName,p=N,f=N)[0]
			
			#Find out if it is a root rigid body and if so, add it to a list
			if len( mel.listRelatives(transformName,p=N,f=N) ) == 0:
				self.roots.append(transformName)
				
			#Attributes
			#Transform
			position = mel.xform(transformName,q=N,ws=N,t=N)
			rotation = mel.xform(transformName,q=N,ws=N,ro=N)

			E = MEulerRotation( radians(rotation[0]), radians(rotation[1]), radians(rotation[2]) )
			Q = E.asQuaternion()

			#Material
			density = mel.getAttr("%s.mass" % shapeName)
			#Shape
			height = 2*mel.getAttr("%s.height" % shapeName)
			width = 2*mel.getAttr("%s.width" % shapeName)
			depth = 2*mel.getAttr("%s.depth" % shapeName)
			
			## The Collision Object
			collision = ode.GeomBox(self.space,lengths=(height,width,depth))
			
			## The Mass and Physics Body
			if density != 0:
				body = ode.Body(self.w)
				mass = ode.Mass()
				mass.setBox(density,height,width,depth)
				body.setMass(mass)
				body.setPosition( position )
				body.setQuaternion((Q.w,Q.x,Q.y,Q.z))
				collision.setBody(body)
			
			## Packaging it up
			a = atom()
			a.collision = collision
			a.body = body
			if density != 0:
				a.mass = mass
			a.name = transformName
			## Adding it to the world class
			self.objects[transformName] = a
			self.dynamics.__dict__[transformName.replace("|","_")[1:]] = a
		
		#Looking for joints which have 1 parent and 1 child capsule
		#Select all joints
		for i in mel.ls(type="joint",l=N):
			#Check that it is dynamic, and if it is, gets is parent and child
			(isDynamic,parent,child) = isDynamicJoint(i)
			
			if isDynamic :
				#The two body objects below are ode objects
				#Get the parent body, if the joint has no parent, then attach it to the world.
				if parent != None:
					parentBody = self.objects[parent].body
				else:
					parentBody = ode.environment
				
				#Get the child body
				childBody = self.objects[child].body
				
				#Get the postion of the joint
				position = mel.xform(i,q=N,ws=N,t=N)
				joint = ode.BallJoint( self.w )
				joint.attach( parentBody, childBody )
				joint.setAnchor( position )
				
				#Packaging it up
				a = atom()
				a.joint = joint
				a.name = i
				
				#Adding it to the world class
				self.joints[i] = a
				if parent == None:
					self.roots.append(i)
				self.dynamics.__dict__[i.replace("|","_")[1:]] = a
		
		#Remove any keyframes which might already exist
		self.cleanup()
	
	def cleanup(self):
		for i in self.objects.keys():
			attrs=[	"translateX","translateY","translateZ",
					"rotateX","rotateY","rotateZ",
					"scaleX","scaleY","scaleZ",
					"visibility"
			]
			for j in attrs:
				attrname = "%s_%s" % (i,j)
				if mel.objExists(attrname):
					mel.delete(attrname)

	def m_subStep(self,name):
		"""This is an internal function, don't use it.
		It recursively travels down the hierarchy of a rig and keys the relevent transforms.
		"""
		if len(name) > 0:
			o = self.objects
			j = self.joints
			
			#Is it a rigid body
			if o.has_key(name) :
				o = self.objects
				pos = o[name].body.getPosition()
				R = o[name].body.getRotation()
				transform = (	R[0], R[3], R[6], 0.,
								R[1], R[4], R[7], 0.,
								R[2], R[5], R[8], 0.,
								pos[0], pos[1], pos[2], 1.0
							)
				try:
					mel.xform(name,ws=N,a=N,matrix=transform)
				except RuntimeError:
					M.displayInfo("Runtime error in Luya. Caused by invalid matrix information. Luya is assuming your rig has exploded.")
				mel.setKeyframe(name)
				
			#Is it a joint
			if j.has_key(name):
				pos = self.joints[name].joint.getAnchor()
				mel.xform(name,ws=N,t=pos)
				mel.setKeyframe(name)
			
			children = mel.listRelatives(name,f=N,c=N,type=["transform","joint"])
			for i in children:
				self.m_subStep(i)
		
	def step(self,timestep=1):
		for i in self.roots:
			self.m_subStep(i)
			
		self.space.collide( self, near_callback )
		self.w.step(timestep)
		self.contacts.empty()
		