#General Modules
import luyaUtils as lu
from mayaUtils import *
import luyaODE as luya ; reload(luya)
from math import pi,degrees,radians
from maya.api import MVector,MQuaternion
from maya.mel import *

## Our Fuzzy Logic Modules
import pyFuzzy; reload(pyFuzzy)
from pyFuzzy import *
from fuzzyRules import interpret

## Given a Current Rotation and and Reference position Compute an Angular Velocity.
class jointTracker:
    def __init__( self, strength=180 ):
        #Define the Angle Input Variable.
        scale  = 90
        u = Stepper(-scale*3, scale*3,[ -3*scale, -scale, scale, 3*scale ])
        self.Angle = InputVariable (u.lower,u.upper,(
            ("Low" , Linear(u(0),u(1)), Linear(u(1),u(2))),
            ("High", Linear(u(1),u(2)), Linear(u(2),u(3)))
        ))
        
        #Define the Anglular Speed Output Value
        u = Stepper(-strength*3, strength*3,[ -3*strength, -strength, strength, 3*strength ])
        self.AngularSpeed = OutputVariable (u.lower,u.upper,(
            ("Low" , Linear(u(0),u(1)), Linear(u(1),u(2))),
            ("High", Linear(u(1),u(2)), Linear(u(2),u(3)))
        ))
        
        #Define and interpret our ruleset
        self.ruleSet = interpret(\
        """
            if Angle is High : AngularSpeed = High
            if Angle is Low : AngularSpeed = Low
        """
        )
        
    ## Do the computation
    def __call__( self, Current, Reference=MVector(0,1,0)):
        Angle = self.Angle ; AngularSpeed = self.AngularSpeed
        
        #Normalize the two input vectors"""
        Reference = Reference.normal()
        Current = Current.normal()
        
        #Compute the rotation axis.
        #The rotation axis flips when Current changes which side of Reference it is on.
        #This is what we use to determin the Direction part of the Angular Velocity.
        Axis = ( Reference ^ Current ).normal()
        
        #Work out the current Error value.
        #This is the angle between Current and Reference
        Error = degrees( Current.angle( Reference ) )
    
        #Fuzzify the Error value
        Angle.fuzz(Error)
        
        #Prepare the output variable for the computation of the Fuzzy Ruleset.
        #Any accumulated membership values are cleared out.
        AngularSpeed.begin()
        
        #Execute the fuzzy rules.
        #This computes an accumulative membership value for each set in the output variable AngularSpeed.
        exec(self.ruleSet)
        
        #Defuzz the Angular velocity.
        #Get a "crisp" value from the AngularSpeed Linguistic Variable.
        Solution = AngularSpeed.defuzz()
        
        #Return thr angular velocity as a Quaternion.
        return MQuaternion( radians( Solution ),Axis)

N= None
halfPi = pi / 2

## Initialise our behaviour
def init( g ):
    #Create an instance of our physics class
    g.w = luya.World( ground = True )
    #Create an instance of our jointTracker class
    g.w.tracker = Tracker.jointTracker()

## Run our Fuzzy Feedback Controller
def run( g ):
    #Loop through each joint in the chain
    for name in g.w.joints.keys():
        #Retrieve the joint
        joint = g.w.joints[name]
        #Retrieve the motor for the joint
        motor = joint.motor
        #Retrieve the current direction of the joint
        Current = Translate(joint.child) - Translate(name)
        #Use our fuzzy joint tracker class to compute an Angular Velocity
        Euler = g.w.tracker( Current ).asEulerRotation()
        
        #Multiply this velocity by a Constant
        kP = 2
        motor.setParam( luya.ParamVel,   kP*Euler.x )
        motor.setParam( luya.ParamVel2, kP*Euler.y )
        motor.setParam( luya.ParamVel3, kP*Euler.z )
        
    #Step the simulation
    g.w.step(0.05)

def end( g ):
    pass
