/// A C++ Implimentation of part of the Python Fuzzy Logic library

#include<cFuzzy.h>

using namespace boost::python ;

/// This function performs a check which raises the GreaterThanError exception
bool checkRange(double lower, double upper){
	if (lower <= upper ){
		return 1 ;
	}else{
		return 0 ;
	}
}

///A fuzzy comparison
double f_is( InputVariable *variable, std::string setname ){
	return variable->f_isEqualTo(setname) ;
}

///A fuzzy if statement
void f_if( double membershipValue, boost::python::list outputVariablesAndSets ){
	int i_len = len( outputVariablesAndSets ) ;
	int i ;
	for( i=0; i < i_len; ++i ) {
		tuple pair = extract<tuple>(outputVariablesAndSets[i]) ;
		OutputVariable *Out = extract<OutputVariable*>(pair[0]) ;
		std::string SetName = extract<std::string>(pair[1]) ;
		
		Out->f_assign( SetName, membershipValue ) ;
	}
}

Curve::Curve(double lower, double upper){
    checkRange(lower,upper) ;
    m_lower = lower ;
    m_upper = upper ;
}

Singleton::Singleton(double value) : Curve( value, value ){
	m_value = value ;
}

double Singleton::compute( double value ) const {
    if (m_value == value) {
        return 1.0 ;
    } else {
        return 0.0 ;
    }
}

Linear::Linear(double lower, double upper) : Curve(lower, upper){
} ;

double Linear::compute( double value ) const {
    return (value - m_lower) / (m_upper - m_lower) ;
}

Sin::Sin(double lower, double upper) : Curve(lower, upper){
} ;

double Sin::compute( double value ) const {
    double l = (value - m_lower) / (m_upper - m_lower);
    return ((sin((l+M_PI/2)*M_PI)/2)+0.5);
}

Set::Set(std::string name, Curve &left, Curve &right){
    m_name = name ;
    checkRange(left.m_upper, right.m_lower);
    
    m_left = &left ;
    m_right = &right ;
}

InputVariable::InputVariable(double lower, double upper, boost::python::tuple sets ){
	checkRange(lower,upper);
	
	m_lower = static_cast<float>(lower) ;
	m_upper = static_cast<float>(upper) ;
	
	int i;
	int len_sets = len( sets ) ;
	
	for ( i=0 ; i < len_sets; ++i ){
		tuple info = extract<tuple>( sets[i] ) ;
		if ( len(info) != 3 ){
			throw std::string("Error, Set Arguments should be: [(\"name\", Curve, Curve ),...] ") ;
		}
		std::string name = extract<std::string>(info[0]) ;
		Curve *left = extract<Curve*>(info[1]) ;
		Curve *right = extract<Curve*>(info[2]);
		
		m_sets[name] = new Set( name, *left, *right );
	}
}

InputVariable::~InputVariable(){
	//MAKE SURE YOU DELETE THE CONTENTS OF m_sets which you are not doing at the moment.
}

void InputVariable::fuzz(double number){
	if (number < m_lower){
		number = m_lower ;
	} else if ( number > m_upper ){
		number = m_upper ;
	}
	m_value = number ;
}

double InputVariable::f_isEqualTo(std::string setname){
	if ( m_sets.find(setname) == m_sets.end() ){
		throw std::string("Error, Invalid Set name.");
	}
	return f_getMembership( *m_sets[setname], m_value ) ;
}

OutputVariable::OutputVariable( double lower, double upper, boost::python::tuple sets ) : InputVariable( lower, upper, sets ) {
	std::map< std::string, Set * >::iterator i = m_sets.begin() ;
	for ( ; i != m_sets.end() ; ++i ){
		m_values[i->first] = 0.0 ;
	}
}

///Reset the output variable's accumulative values to 0.0 .
void OutputVariable::begin(){
	m_values.erase( m_values.begin(), m_values.end() ) ;
	std::map< std::string, Set * >::iterator i = m_sets.begin() ;
	for ( ; i != m_sets.end() ; ++i ){
		m_values[i->first] = 0.0 ;
	}
}

double OutputVariable::defuzz( boost::python::tuple rez, std::string filename ){
	return defuzzCentroid( rez, filename ) ;
}

///Centroid deffuzzification
double OutputVariable::defuzzCentroid( boost::python::tuple rez, std::string filename ){
        ///Specify what resolution the image should be
        int width = extract<int>(rez[0]) ;
        int height = extract<int>(rez[1]) ;
        
        double d_width = extract<double>(rez[0]) ;
        double d_height = extract<double>(rez[1]) ;
        
        int pixbuff[width][height] ;
        
        ///Get the variables range
        double vrange = m_upper-m_lower ;
        double prange = d_width ;
        
        ///    cX and cY are the coordinates of the centroid.
        ///   sum will be used to compute the values of cX and cY.
        int cX = 0	;
        int cY = 0	;
        double sum = 0	;
                
        ///Compute the centroid
        int i;
        for ( i=0; i<width; ++i ){
            
			//	Map the position of the the pixel with x coordinate i to a coordinate
			//	space of the output variable.
			//	k = i / prange gives a value along the width of the pixbuf from 0 to 1
			//	k*vrange + lower maps k to the output variable.

            double d_k = static_cast<double>(i)/prange ;
            double j = d_k * vrange + m_lower ;
                
			//	Compute the highest y coordinate for the coordinate j.
			//	The point on the curve.
            double d_y = 0 ;

            ///Loop through each set in the output variable
            std::map< std::string, Set * >::iterator k = m_sets.begin() ;
			for ( ; k != m_sets.end() ; ++k ){

				//	Get the membership value of j in set k described by the curves
				//	which define it.
                double ny = f_getMembership( *k->second, j ) ;
                
				//	Check for membership value defined by fuzzy operations on the variable.
				//	Compare the membership values and take the lowest. This does the chopping off
				//	the top of the set.
                if ( m_values.find( k->first ) != m_values.end() ){
                    ny = f_and(ny, m_values[ k->first ] ) ;
                }
                
				//	Compare the computed membership value with any from overlapping sets and take
				//	the highest.
                d_y = f_or(d_y,ny) ;
			}

            int i_y = static_cast<int>( d_y * d_height ) ;
            for ( int m=0 ; m < i_y; ++m ){
                pixbuff[m][i] = 1 ;
                cX += i ;
                cY += m ;
                sum+=1 ;
            }
        }

        if ( sum == 0) {
            cX = width / 2  ;
            cY = height / 2 ;
        } else {
            cX /= static_cast<int>(sum) ;
            cY /= static_cast<int>(sum) ;
        }
        double d_cX = static_cast<double>(cX) ;
        
        ///The two lines below are for debugging purposes only. They write the contents of the pixbuff to an xpm file.
        if (filename.size() > 0 ){
            ///markPoint(pixbuff,cX,cY) ;
            ///pixbuffToXPM(pixbuff,width,height,filename=filename) ;
        }
        
        return ((d_cX / prange) *vrange) + m_lower ;
}

void OutputVariable::f_assign(std::string setname, double newValue){
	if ( m_sets.find(setname) == m_sets.end() ) {
		throw std::string("Set") + setname  + std::string(" does not exist, so you can't assign a value to it");
	}
	if ( newValue < 0 and newValue > 1){
		throw std::string("Membership values must be between 0 and 1");
	}

	//	Take whichever value is greatest between newVale and the currentValue.
	//	currentValue is the last greatest membership value.
	//	If no membership value existed previously then take newValue.
	
	double currentValue ;
	
	if ( m_values.find(setname) != m_values.end() ){
		currentValue = m_values[setname] ;
	}else{
		currentValue = newValue ;
	}
	m_values[setname] = f_or(currentValue,newValue) ;
}


///Mark a point on the pixbuff.
///For debugging purposes.
void markPoint(int *buff[], int x, int y){
	#if 0
    """Debug function,
        Draws a cross on the defussification buffer.
    """
    w = 2
    h = 2
    for i in range(x-w,x+w+1):
        buff[y][i] = 2
    for i in range(y-h,y+h+1):
        buff[i][x] = 2
    #endif
}

///Draw the contents of the pixelbuffer to a file
///For debugging purposes.
void pixbuffToXPM(int *buff[], int width, int height, std::string filename ){
	#if 0
    output = open(filename,"w")
    outstr = \
"""
/* XPM */
static char out_xpm[] = { \"%s %s 3 1\",
                        "   c #FFFFFF",
                        ".  c #000000",
                        "+  c #FF0000"
""" % (width,height)

    for y in range(height):
        outstr += ',"'
        for x in range(width):
            j = buff[(y*-1)+height-1][x]
            if j == 0:
                outstr += " "
            elif j == 1:
                outstr += "."
            else:
                outstr += "+"
        outstr += '"\n'

    outstr += "};\n"

    output.write(outstr)
    output.close()
    #endif
}

///Fuzzy Intersection
double f_and( double m1, double m2){
    if (m1 < m2){
        return m1 ;
	}else {
        return m2 ;
	}
}

///Fuzzy Union
double f_or( double m1, double m2){
    if (m1 > m2){
        return m1 ;
    }else {
        return m2 ;
    }
}

///Fuzzy inversion
double f_not(double m){
    return 1.0 - m ;
}

///Get the membership value of the variable of < variable > in the Set set.
double f_getMembership( Set &set, double value ){
    Curve* left = set.m_left ;
    Curve* right = set.m_right;
    
    if (value > left->m_lower and value < left->m_upper) {
        ///Are we on the left slope
        double v = left->compute( value ) ;
        return v ;
        
    } else if ( value > right->m_lower and value < right->m_upper) {
        ///Are we on the right slope?
        return 1 - right->compute( value ) ;
        
    } else if ( value <= right->m_lower and value >= left->m_upper) {
        ///Are we in the middle
        return 1.0 ;
        
    }else{
    	return 0.0 ;
    }
}


